import logging

import requests

from sepiida.requests import privileged_session, user_session

LOGGER = logging.getLogger(__name__)

CONFIG = {
  'PAO_ROOT': None,
}
def configure(root):
    CONFIG['PAO_ROOT'] = root

class MembershipsBadRequestError(Exception):
    pass

class MembershipsUnauthorizedError(Exception):
    pass

class MembershipsForbiddenError(Exception):
    pass

class MembershipsNotFoundError(Exception):
    pass

class MembershipsRequestError(Exception):
    pass

class MembershipsConnectionError(Exception):
    pass

class MembershipsPaoRootError(Exception):
    pass

def is_privileged():
    return is_privileged.nest > 0
is_privileged.nest = 0

def _send_request(http_method, uri=None, query=None, payload=None, timeout=10, privileged=False):
    privileged = is_privileged() or privileged
    pao_root = CONFIG.get('PAO_ROOT')
    if not pao_root:
        raise MembershipsPaoRootError(
            "You must include a value for PAO_ROOT in your app config,"
            "such as https://pao.service, before querying the membershipss system"
        )

    url = uri or '{}/memberships/'.format(pao_root)

    session = privileged_session() if privileged else user_session()
    try:
        if query:
            response = session.request(http_method, url, data=query, timeout=timeout)
        else:
            response = session.request(http_method, url, json=payload, timeout=timeout)
    except (requests.exceptions.ReadTimeout, requests.exceptions.ConnectionError) as e:
        raise MembershipsConnectionError(e)

    if not response.ok:
        LOGGER.debug('Error sending request %s %s with payload %s: %s %s',
            http_method,
            url,
            payload,
            response.status_code,
            response.text,
        )
        message = "Failed Request, Response: {}, {}, Payload: {}".format(response.status_code, response.text, payload)
        if response.status_code == 400:
            raise MembershipsBadRequestError(message)
        elif response.status_code == 401:
            raise MembershipsUnauthorizedError(message)
        elif response.status_code == 403:
            raise MembershipsForbiddenError(message)
        elif response.status_code == 404:
            raise MembershipsNotFoundError(message)
        else:
            raise MembershipsRequestError(message)

    return response

def _payload(group, user):
    return {
        'group'     : group,
        'user'      : user,
    }

def create(group, user, privileged=False):
    http_method = 'POST'
    payload = _payload(group, user)
    response = _send_request(http_method, payload=payload, privileged=privileged)
    return response.headers['Location']

def delete(uri, privileged=False):
    http_method = 'DELETE'
    response = _send_request(http_method, uri=uri, privileged=privileged)
    assert response.status_code == 204

def get(uri=None, privileged=False):
    http_method = 'GET'
    response = _send_request(http_method, uri=uri, privileged=privileged)
    return response.json()

def search(group=None, user=None, privileged=False):
    http_method = 'GET'
    filters = []
    if group:
        filters.append("filter[group]={}".format(group))
    if user:
        filters.append("filter[user]={}".format(user))
    filters_qs = "&".join(filters)
    response = _send_request(http_method, query=filters_qs, privileged=privileged)
    return response.json()['resources']



import contextlib
import logging
import logging.config
import os
import time

FORMAT = '[%(asctime)s] %(levelname)s pid:%(process)d %(name)s:%(lineno)d %(message)s'
LOGGER = logging.getLogger(__name__)

def level_for_environment():
    return logging.INFO if os.environ.get('ENVIRONMENT') == 'prod' else logging.DEBUG

def setup_logging(level=None):
    level = level if level else level_for_environment()
    configuration = {
        'version'   : 1,
        'disable_existing_loggers'  : False,
        'formatters'                : {
            'standard'              : {
                'format'            : FORMAT,
                'dateformat'        : '%d/%b/%Y:%H:%M:%S %z',
            },
        },
        'handlers'                  : {
            'console'               : {
                'level'             : level,
                'class'             : 'logging.StreamHandler',
                'formatter'         : 'standard',
            },
        },
        'loggers'                   : {
            ''                      : {
                'handlers'          : ['console'],
                'level'             : level,
                'propogate'         : True,
            },
        },
    }
    logging.config.dictConfig(configuration)

@contextlib.contextmanager
def warning_timer(seconds, message, start=None):
    start = start or time.time()
    yield
    end = time.time()
    if (end - start) > seconds:
        LOGGER.warning("Took %s seconds doing %s", end - start, message)



import logging
import uuid

import flask
import requests
from flask_httpauth import HTTPBasicAuth

from sepiida.errors import APIException, AuthenticationException, Error

try:
    import uwsgi
except ImportError:
    uwsgi = None

auth = HTTPBasicAuth()
LOGGER = logging.getLogger(__name__)

def register_session_handlers(app, pao_root, whitelist=None, internal_domains=None):
    app.config['SEPIIDA_PAO_ROOT'] = pao_root
    app.config['SEPIIDA_INTERNAL_DOMAINS'] = app.config.get('SEPIIDA_INTERNAL_DOMAINS', []) + list(internal_domains or [])

    app.before_request(require_user)
    whitelist = whitelist or []
    for endpoint in whitelist:
        public_endpoint(app, endpoint)

    app.config['PUBLIC_ENDPOINTS'] = app.config.get('PUBLIC_ENDPOINTS', []) + whitelist

    @auth.verify_password
    def verify_password(username, password): # pylint: disable=unused-variable
        if username == 'api' and password == flask.current_app.config.get('API_TOKEN', None):
            return True
        else:
            url = '{}/sessions/'.format(pao_root)
            response = requests.get(url, auth=(username, password))
            if not response.status_code == 200:
                return False
            flask.g.current_user = response.json()
            return True

def add_origin_handler(app):
    def extract_origin():
        if uwsgi:
            uwsgi.set_logvar('origin', flask.request.headers.get('Origin', 'none'))
    app.before_request(extract_origin)

def add_request_id_handler(app):
    def extract_request_id():
        flask.g.request_id = flask.request.headers.get('Request-ID', str(uuid.uuid4()))
        if uwsgi:
            uwsgi.set_logvar('request_id', flask.g.request_id)
    def add_request_id(response):
        response.headers['Request-ID'] = current_request_id()
        return response
    app.before_request(extract_request_id)
    app.after_request(add_request_id)

def current_request_id():
    if flask.has_request_context():
        return flask.g.get('request_id', 'none')
    else:
        return 'backend'

def current_user_uri():
    try:
        return flask.g.current_user.get('uri', None)
    except AttributeError:
        return None

def current_user():
    if flask.g.endpoint_authenticated:
        return flask.g.current_user

    if 'uri' not in flask.g.current_user:
        return

    if 'uuid' in flask.g.current_user:
        return flask.g.current_user

    url = '{}/sessions/'.format(flask.current_app.config['SEPIIDA_PAO_ROOT'])
    cookies = { 'session': flask.request.cookies.get('session', None) }
    response = requests.get(url, cookies=cookies)
    if not response.status_code == 200:
        raise AuthenticationException(
                status_code=400,
                error_code='retrieving-user-details-error',
                title="An error occured while retrieving the user details.")
    flask.g.current_user = response.json()
    return flask.g.current_user


def is_whitelisted(endpoint, method):
    blueprint, _, basename = endpoint.rpartition('.')
    basename = basename or blueprint

    endpoint_with_method = '{}.{}'.format(endpoint, method.lower())
    defaults = flask.request.url_rule.defaults
    if defaults and not defaults.get('_single_resource', False) and method == "GET":
        endpoint_with_method = '{}.list'.format(endpoint)

    public_endpoints = flask.current_app.config['PUBLIC_ENDPOINTS']
    return any([
        basename             in public_endpoints,
        endpoint             in public_endpoints,
        endpoint_with_method in public_endpoints,
    ])


def public_endpoint(app, f):
    try:
        endpoint_name = f.__name__
    except AttributeError:
        endpoint_name = f
    app.config['PUBLIC_ENDPOINTS'] = app.config.get('PUBLIC_ENDPOINTS', []) + [endpoint_name]

def get_endpoint_authentication():
    resource = getattr(flask.current_app, 'endpoints', {}).get(flask.request.endpoint, None)
    if resource:
        try:
            return resource.authenticate()
        except Error as e:
            raise APIException(errors=[e])

def require_user():
    # pylint: disable=too-many-branches
    flask.g.current_user = {}
    flask.g.endpoint_authenticated = False
    if 'Access-Control-Request-Method' in flask.request.headers and flask.request.method == 'OPTIONS':
        return

    elif not flask.request.endpoint:
        return

    elif flask.request.path.startswith(flask.current_app.static_url_path):
        return

    elif is_whitelisted(flask.request.endpoint, flask.request.method):
        return

    endpoint_authentication = get_endpoint_authentication()
    if endpoint_authentication:
        flask.g.current_user = endpoint_authentication
        flask.g.endpoint_authenticated = True
    elif flask.request.authorization:
        if auth.authenticate(flask.request.authorization, flask.request.authorization.password):
            return
        else:
            raise AuthenticationException(
                    status_code=403,
                    error_code='invalid-credentials',
                    title='The API key and secret you provided were not recognized')

    elif 'session' not in flask.request.cookies and 'Authorization' not in flask.request.headers:
        raise AuthenticationException(
                status_code=403,
                error_code='no-credentials-provided',
                title="You must include a valid session cookie or an API token and secret "
                    "encoded in your Authorization header according to Basic Auth standards (RFC 2617)")

    elif 'Authorization' in flask.request.headers:
        raise AuthenticationException(
                status_code=400,
                error_code='not-well-formed-authorization-header',
                title="The authorization header you have provided was not properly formatted. "
                    "Please see RFC 2617")

    elif len(flask.session) == 0:
        raise AuthenticationException(
                status_code=403,
                error_code='empty-session-provided',
                title="The session cookie you have provided does not contain any data")

    elif 'user_uri' not in flask.session:
        raise AuthenticationException(
                status_code=403,
                error_code='invalid-session-provided',
                title="The session cookie you have provided does not contain valid data")

    elif 'user_uri' in flask.session:
        flask.g.current_user['uri'] = flask.session['user_uri']

        if not flask.g.current_user['uri']:
            raise AuthenticationException(
                    status_code=400,
                    error_code='unkown-user-error',
                    title="The user information you have provided is not valid or does not exist.")

    else:
        raise AuthenticationException(
                status_code=403,
                error_code='unkown-authentication-error',
                title="Not sure what happened, but authentication failed. Tell Authentise")

def is_api_user():
    if flask.request.authorization:
        password = auth.get_password_callback(flask.request.authorization.username)
        return auth.authenticate(flask.request.authorization, password)
    else:
        return False

def is_internal_user():
    return bool(flask.request.authorization) and \
           flask.request.authorization.username == 'api' and \
           flask.request.authorization.password == flask.current_app.config.get('API_TOKEN')



import json
import logging
from urllib.parse import urljoin

import requests

LOGGER = logging.getLogger(__name__)

def handle(service_root, url, method, status, uri):
    if url is None:
        LOGGER.info("Not doing callback, no URL provided")
        return

    if method is None:
        LOGGER.info("Not doing callback, no method provided")
        return

    payload = {
        'status'    : status,
        'uri'       : uri,
    }

    callback_payload = {
        "method" : method,
        "url"    : url,
        "body"   : json.dumps(payload),
    }
    response = requests.post(urljoin(service_root, '/job/'), json=callback_payload)
    if not response.ok:
        LOGGER.error("Failed to put callback job on the callback queue: %s %s", response.status_code, response.text)



import json
import logging

from werkzeug.exceptions import HTTPException

LOGGER = logging.getLogger(__name__)

class SepiidaError(Exception):
    pass

class AppError(SepiidaError):
    pass

class RoutingError(SepiidaError):
    pass

class Specification():
    """
    This class is in charge of identifying the errors that an APIEndpoint may emit. They act as a sort of translation between
    internal error types and HTTP status codes with error title and description. If a status_code, error_code, title or details
    are provided then they will be used any time an exception is raised inside the request handler that matches the type of
    this specification.

    The `docs` get extracted from this class when producing RAML documentation on the errors an endpoint may emit
    """
    def __init__(self, exception_class, status_code=None, error_code=None, title=None, details=None, docs=None):
        self.details            = details
        self.docs               = docs
        self.error_code         = error_code
        self.exception_class    = exception_class
        self.status_code        = status_code
        self.title              = title

    def translate(self, exception):
        if not isinstance(exception, self.exception_class):
            raise TypeError("Cannot translate {}, it isn't an instance of {}".format(exception, self.exception_class))

        details     = self.details if self.details else getattr(exception, 'details', None)
        error_code  = self.error_code if self.error_code else getattr(exception, 'error_code', self.exception_class.__name__)
        status_code = self.status_code if self.status_code else getattr(exception, 'status_code', 400)
        title       = self.title if self.title else getattr(exception, 'title', str(exception))
        return Error(
            details     = details,
            error_code  = error_code,
            status_code = status_code,
            title       = title,
        )

    def __str__(self):
        return "Error Specification {} -> {} ({})".format(self.exception_class, self.status_code, self.error_code)

    def __repr__(self):
        return str(self)

class Error(SepiidaError):
    def __init__(self, error_code, title, details=None, status_code=400):
        super().__init__()
        self.details        = details
        self.error_code     = error_code
        self.status_code    = status_code
        self.title          = title

    def to_dict(self):
        result = {
            'code'  : self.error_code,
            'title' : self.title,
        }
        if self.details is not None:
            result['details'] = self.details
        return result

class APIException(HTTPException):
    def __init__(self, errors, status_code=None, headers=None):
        super().__init__()
        self.errors         = errors
        self.headers        = headers if headers is not None else {'Content-Type': 'application/json'}
        self._status_code   = status_code

    @property
    def code(self):
        return self._status_code if self._status_code is not None else self.errors[0].status_code

    @property
    def data(self):
        return {'errors': [e.to_dict() for e in sorted(self.errors, key=lambda x: x.error_code + x.title)]}

    def get_body(self, environ=None):
        return json.dumps(self.data)

    def get_headers(self, environ=None):
        return self.headers

class AuthenticationException(HTTPException):
    def __init__(self, status_code=403, error_code='unauthorized', title='You are not authorized'):
        super().__init__()
        self.code       = status_code
        self.error_code = error_code
        self.title      = title

    def get_body(self, environ=None):
        return json.dumps({
            'errors'    : [{
                'code'  : self.error_code,
                'title' : self.title,
            }]
        })

    def get_headers(self, environ=None):
        return {'WWW-Authenticate': 'Custom realm="Authentise"'}

def api_error(title="An error occurred", status_code=400, error_code='error'):
    # Translate from the outer names so that the inner names can shadow without
    # losing data and without having to use a different parameter name at each level
    defaults = {
        'error_code'    : error_code,
        'status_code'   : status_code,
        'title'         : title,
    }
    def __init__(self, title=None, status_code=None, error_code=None): # pylint: disable=unused-argument
        my_kwargs = defaults.copy()
        for prop in ['title', 'status_code', 'error_code']:
            val = locals()[prop]
            if val is not None:
                my_kwargs[prop] = val
        Error.__init__(self, **my_kwargs)
    name = 'generated-exception-class'
    return type(name, (Error,), {'__init__': __init__})


EmptyPayload        = api_error(status_code=400, error_code='empty-payload')
InvalidPayload      = api_error(status_code=400, error_code='invalid-payload')
ResourceNotFound    = api_error(status_code=404, error_code="resource-not-found")
Unauthorized        = api_error(status_code=403, error_code="Unauthorized")
UnableToSerialize   = api_error(status_code=400,
    error_code  = 'unable-to-serialize',
    title       = ("No serializers are able to respond in a format you would accept. "
                   "Check your Accept header and alter it to accept more types for this request")
)



import base64
import contextlib
import io
import logging
import urllib.parse

import requests

import sepiida.config
import sepiida.zipbuf


LOGGER = logging.getLogger(__name__)

class StorageException(Exception):
    pass

class AlreadyUploadedException(StorageException):
    pass

class NotUploadedException(StorageException):
    pass

class NoFileException(StorageException):
    pass

def _auth_header():
    config = sepiida.config.get()
    credentials = 'api:{}'.format(config.sepiida.api_token).encode('utf-8')
    return {'Authorization' : 'Basic {}'.format(base64.b64encode(credentials).decode('utf-8'))}

def upload_link(key, bucket, overwrite=False):
    config = sepiida.config.get()
    payload = {
        'bucket'    : bucket,
        'key'       : str(key),
    }
    url = urllib.parse.urljoin(config.sepiida.storage, 'file/')

    if len('{}?filter[key]={}'.format(url, key)) <= 2000:
        response = requests.get('{}?filter[key]={}'.format(url, key), headers=_auth_header())
    else:
        response = requests.get(url, data='filter[key]={}'.format(key), headers=_auth_header())

    if not response.ok:
        raise StorageException('Unknown storage error: {}'.format(response.text))

    resources = response.json()['resources']
    if len(resources) < 1:
        response = requests.post(url, json=payload, headers=_auth_header())
        if not response.ok:
            try:
                errors = response.json()['errors']
                if 'DuplicateKeyError' in [error['code'] for error in errors]:
                    raise AlreadyUploadedException('A file with key {} has already been uploaded'.format(key))
            except (ValueError, KeyError):
                pass
            raise StorageException('Unknown storage error: {}'.format(response.text))
        upload_location = response.headers['upload-location']
    else:
        if resources[0].get('uploaded') and not overwrite:
            raise AlreadyUploadedException('A file with key {} has already been uploaded'.format(key))

        upload_location = resources[0].get('upload-location')

    return upload_location

def download_link(key):
    config = sepiida.config.get()
    url = urllib.parse.urljoin(config.sepiida.storage, 'file/')

    if len('{}?filter[key]={}'.format(url, key)) <= 2000:
        response = requests.get('{}?filter[key]={}'.format(url, key), headers=_auth_header())
    else:
        response = requests.get(url, data='filter[key]={}'.format(key), headers=_auth_header())

    if not response.ok:
        raise StorageException('Unknown storage error: {}'.format(response.text))

    resources = response.json()['resources']
    if len(resources) < 1:
        raise NoFileException('A file with key {} does not exist in the database'.format(key))

    download_location = resources[0].get('content')

    if not download_location:
        raise NotUploadedException('A file with key {} has not yet been uploaded'.format(key))

    return download_location


def get_files(keys):
    config = sepiida.config.get()
    url = urllib.parse.urljoin(config.sepiida.storage, 'file/')
    _filter = ','.join(map(str, keys))

    response = requests.get(url, data='filter[key]={}'.format(_filter), headers=_auth_header())
    if not response.ok:
        raise StorageException('Unknown storage error: {}'.format(response.text))

    resources = response.json()['resources']
    file_to_key_map = {resource['key'] : resource for resource in resources}

    return file_to_key_map

def put(key, bucket, content, mimetype, overwrite=False, compress=False):
    upload_location = upload_link(key, bucket, overwrite)

    if compress:
        if isinstance(content, bytes):
            content = io.BytesIO(content)
        content = sepiida.zipbuf.ZipBuffer(content) # pylint: disable=redefined-variable-type
        mimetype = 'application/zlib.' + mimetype
        LOGGER.debug("Adding compression to data and changing mimetype to %s", mimetype)

    response = requests.put(upload_location, data=content, headers={'content-type': mimetype})
    if not response.ok:
        raise StorageException('Unknown storage error: {}'.format(response.text))

def get(key, output_filename=None):
    download_location = download_link(key)

    response = requests.get(download_location, stream=True)
    if not response.ok:
        raise StorageException('Unknown storage error: {}'.format(response.text))

    stream = response.raw
    content_type = response.headers.get('content-type')
    if content_type.startswith('application/zlib.'):
        LOGGER.debug("Found content type of %s, unzipping", content_type)
        stream = sepiida.zipbuf.UnzipBuffer(stream)

    if output_filename:
        with open(output_filename, 'wb') as f:
            while True:
                block = stream.read(1024 * 100)
                if not block:
                    break
                f.write(block)
        stream = open(output_filename, 'rb')
    return contextlib.closing(stream)



import base64
import collections
import contextlib
import json as JSON
import logging
import re
import urllib.parse
import zlib
from uuid import uuid4

import httpretty as HTTPretty
import itsdangerous
import pytest
import werkzeug.urls

import sepiida.backend
import sepiida.config
import sepiida.jwt
import sepiida.parsing
import sepiida.permissions
import sepiida.requests

LOGGER = logging.getLogger(__name__)

class JSONClient():
    def __init__(self, client):
        self.client = client

    def _do_send(self, *args, method, json, **kwargs):
        if 'data' in kwargs:
            raise Exception("Don't specify a data parameter when using the JSONClient")
        data = JSON.dumps(json)
        headers = kwargs.pop('headers', {})
        headers = headers or {}
        headers['Content-Type'] = 'application/json'
        headers['Accept'] = 'application/json'

        auth = kwargs.pop('auth', ())
        if auth:
            headers['Authorization'] = _basic_auth(*auth)

        jwt = kwargs.pop('jwt', None)
        if jwt:
            headers['jwt'] = sepiida.jwt.encode(payload=json, **jwt)

        method = getattr(self.client, method)
        return method(*args, data=data, headers=headers, **kwargs)

    def post(self, *args, json, **kwargs):
        return self._do_send(*args, json=json, method='post', **kwargs)

    def put(self, *args, json, **kwargs):
        return self._do_send(*args, json=json, method='put', **kwargs)

    def get(self, *args, **kwargs):
        headers = kwargs.pop('headers', {})
        headers = headers or {}
        auth = kwargs.pop('auth', ())
        if auth:
            headers['Authorization'] = _basic_auth(*auth)
        headers['Accept'] = 'application/json'

        jwt = kwargs.pop('jwt', None)
        if jwt:
            headers['jwt'] = sepiida.jwt.encode(payload='', **jwt)
        return self.client.get(*args, headers=headers, **kwargs)

    def options(self, *args, **kwargs):
        headers = kwargs.pop('headers', {})
        headers = headers or {}
        headers['Content-Type'] = 'application/json'
        return self.client.options(*args, headers=headers, **kwargs)

    def __getattr__(self, name):
        return getattr(self.client, name)

@pytest.yield_fixture
def json_client(app):
    with app.test_client() as client:
        yield JSONClient(client)

@pytest.fixture
def make_session(app):
    @contextlib.contextmanager
    def _inner(user_uri):
        with app.test_client() as client:
            with client.session_transaction() as session:
                session['user_uri'] = user_uri
            yield JSONClient(client)
    return _inner

@pytest.yield_fixture
def json_client_session(make_session, user): # pylint: disable=redefined-outer-name
    with make_session(user['uri']) as session:
        yield session

@pytest.yield_fixture
def httpretty():
    import socket
    old_socket_type = socket.SocketType
    HTTPretty.enable()
    yield HTTPretty
    HTTPretty.disable()
    socket.SocketType = old_socket_type
    HTTPretty.reset()

def _basic_auth(username, password):
    credentials = '{}:{}'.format(username, password).encode('utf-8')
    return 'Basic {}'.format(base64.b64encode(credentials).decode('utf-8'))

@pytest.fixture()
def basic_auth():
    return _basic_auth

# This has to be named 'configuration' because pytest_flask is already providing a (mostly useless)
# 'config' fixture that conflicts with this fixture
@pytest.fixture(scope='session', autouse=True)
def configuration(config_specification): # pylint: disable=redefined-outer-name
    values = sepiida.config.extract_test_values(config_specification)
    sepiida.config.get.cached = values
    return values

@pytest.fixture
def user():
    uuid = str(uuid4())
    return {
        'name'          : 'Kenny Powers',
        'password'      : 'kpow-was-here',
        'uri'           : 'https://users.service/users/{}'.format(uuid),
        'username'      : 'kpowers',
        'uuid'          : uuid,
        'emails'        : [{
            'uuid'      : str(uuid4()),
            'user_uuid' : uuid,
            'email'     : 'kenny@powers.com',
        }],
    }

@pytest.fixture
def pao_user(configuration, httpretty, user): # pylint: disable=redefined-outer-name
    httpretty.register_uri(
        httpretty.GET,
        configuration.pao_root + '/sessions/',
        body=JSON.dumps(user),
        status=200
    )

    httpretty.register_uri(
        httpretty.GET,
        user['uri'],
        body=JSON.dumps(user),
        status=200
    )

    return user

def get_cookie_value(request, name):
    headers = [value for name, value in request.headers.items() if name == 'Cookie']
    headers = [value for value in headers if value.startswith(name + '=')]
    name, _, value = headers[-1].partition('=') if headers else (None, None, None)
    return value

def get_session_data(app, request):
    serializer = app.session_interface.get_signing_serializer(app)
    session_data = get_cookie_value(request, 'session')
    session = serializer.loads(session_data) if session_data else {}
    return session

def _get_session(headers):
    for key, value in headers.items():
        if key == 'Cookie' and value.startswith('session='):
            encoded = value[len('session='):]
            compressed = encoded.startswith('.')
            encoded = encoded[1:] if compressed else encoded
            content = encoded.split('.')[0]
            data = itsdangerous.base64_decode(content)
            raw = zlib.decompress(data) if compressed else data
            session = JSON.loads(raw.decode('utf-8'))
            return session
    return {}

def _extract_filters(request):
    body = request.body
    if not (body or request.querystring):
        return {}
    if request.body:
        search = werkzeug.urls.url_decode(body, charset='utf-8', errors='strict')
        filters = [sepiida.parsing.FilterArgument(k, [v]) for k, v in search.items() if k.startswith('filter')]
    else:
        search = request.querystring
        filters = [sepiida.parsing.FilterArgument(k, v) for k, v in search.items() if k.startswith('filter')]
    return filters

def _filter_by_holder(grants, session):
    # Apply implicit permissions filter. This is to mimic the fact that a
    # user can only list permissions that grants them something, either by
    # belonging to a group or by being themselves
    # This is not perfectly accurate - we don't mimic group behavior
    # here. Sorry.
    user_uri = session.get('user_uri')
    if user_uri:
        LOGGER.debug("Filtering out grants whose holder or creator are not %s", user_uri)
        grants = [g for g in grants if (g.holder == user_uri or g.created_by == user_uri or g.holder in (None, '*'))]
    return grants

@pytest.yield_fixture
def permission(app, httpretty): # pylint: disable=redefined-outer-name
    Grant = collections.namedtuple(typename='Grant', field_names=[
        'created_by',
        'holder',
        'namespace',
        'object',
        'resource_name',
        'right',
        'uri'
    ])
    root = sepiida.permissions.CONFIG['PAO_ROOT']
    assert root, "You need to set up your PAO_ROOT in your app setup"

    class Permission():
        def __init__(self):
            self.creations  = {}
            self.deletions  = set()
            self.grants     = {}
            httpretty.register_uri(
                httpretty.POST,
                '{}/permissions/'.format(root),
                body=self.capture_post,
            )
            httpretty.register_uri(
                httpretty.DELETE,
                re.compile(root + '/permissions/[\\-\\w]+/'),
                body=self.capture_delete,
            )
            httpretty.register_uri(
                httpretty.GET,
                '{}/permissions/'.format(root),
                body=self.capture_list,
            )
            httpretty.register_uri(
                httpretty.GET,
                re.compile(root + '/permissions/[\\-\\w]+/'),
                body=self.capture_get,
            )

        def matching_grants(self, filters, session):
            grants = [g for g in self.grants.values()]
            LOGGER.debug("Starting with %d permission grants available", len(grants))
            for f in filters:
                grants = [g for g in grants if getattr(g, f.name) in f.values]
                LOGGER.debug("%d permission grants left after filtering '%s in %s'", len(grants), f.name, f.values)
            grants = _filter_by_holder(grants, session)
            LOGGER.debug("Found %d grants matching search query %s", len(grants), filters)
            return grants

        def capture_get(self, request, uri, headers):
            headers['Content-Type'] = 'application/json'
            session = get_session_data(app, request)
            grant = self.grants.get(uri)
            grant = _filter_by_holder([grant], session) if grant else None
            grant = grant[0] if grant else None
            LOGGER.debug("Captured request for %s and found %s", uri, grant)
            if not grant:
                return 404, headers, ''
            body = {
                'uri'           : grant.uri,
                'holder'        : grant.holder,
                'namespace'     : grant.namespace,
                'object'        : grant.object,
                'resource_name' : grant.resource_name,
                'right'         : grant.right,
                'created_by'    : grant.created_by,
            }
            return 200, headers, JSON.dumps(body)

        def capture_list(self, request, dummy_uri, headers):
            headers['Content-Type'] = 'application/json'
            filters = _extract_filters(request)
            session = get_session_data(app, request)
            matching_grants = self.matching_grants(filters, session)
            body = {'resources': [{
                'uri'           : grant.uri,
                'holder'        : grant.holder,
                'namespace'     : grant.namespace,
                'object'        : grant.object,
                'resource_name' : grant.resource_name,
                'right'         : grant.right,
                'created_by'    : grant.created_by,
            } for grant in matching_grants]}
            LOGGER.debug("Captured request to list permissions and returning %d permission grants", len(matching_grants))
            return 200, headers, JSON.dumps(body)

        def capture_delete(self, dummy_request, uri, dummy_headers):
            LOGGER.debug("Captured request to delete permission %s", uri)
            self.deletions.add(uri)
            if uri in self.grants:
                del self.grants[uri]
                LOGGER.debug("Removed %s from the list of available permission grants", uri)
                return 204, {}, ''
            else:
                return 404, {}, ''

        def capture_post(self, request, dummy_uri, headers):
            privileged = _uses_privileged_auth(request.headers['Authorization'])
            session = _get_session(request.headers)
            body = JSON.loads(request.body.decode('utf-8'))
            body['privileged'] = privileged
            uuid = uuid4()
            location = '{}/permissions/{}/'.format(root, uuid)
            headers['Location'] = location
            self.creations[location] = body
            self.grants[location] = Grant(
                created_by      = session.get('user_uri', 'unknown'),
                holder          = body['holder'],
                namespace       = body['namespace'],
                resource_name   = body.get('resource_name'),
                right           = body['right'],
                object          = body['object'],
                uri             = location,
            )
            LOGGER.debug("Created new fake permission %s", self.grants[location])
            return 201, headers, ''

        def created(self, object_, holder, right, namespace, resource_name, privileged=False):
            expected = {
                'holder'        : holder,
                'namespace'     : namespace,
                'object'        : object_,
                'privileged'    : privileged,
                'resource_name' : resource_name,
                'right'         : right,
            }
            return expected in self.creations.values()

        def deleted(self, uri):
            return uri in self.deletions

        def grant(self, holder, namespace, right, object_, resource_name, created_by='unknown'):
            uuid = uuid4()
            holder = holder if holder is not None else '*'
            uri = '{}/permissions/{}/'.format(root, uuid)
            self.grants[uri] = Grant(
                created_by      = created_by,
                holder          = holder,
                namespace       = namespace,
                object          = object_,
                resource_name   = resource_name,
                right           = str(right),
                uri             = uri,
            )
            LOGGER.debug("Directly creating permission grant %s", self.grants[uri])
            return uri

    yield Permission()

def api(app):
    LOGGER.warning("You are using the api fixture. Use the app fixture instead")
    return app

@pytest.fixture(scope="session")
def external_cert():
    return """-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAqvZ8i1qxb5ijA6Z/oS1D
U9jZM4ZXRLEbErOt541cZC4/AbFjxRfc4vTNWux9JYIjRQ761XJnPNzRvOMnh2Dm
3s8O1MVC2QS4TqRz3ga2QjmqA9ISLmV3h01K9EqeCVlYLjR0yMpUaSJcwVD7kXJC
kATl4svjjpSlKNu/THd8Ybzq03WgMTB5SwysIWXO4jREMqvXGgxVGs54VLullTZy
psWvsm7GOX+wZfTiGk3YoWCI0xYBKWD5IRn5ykk1vt6Hr8m3qMLN8Hjn+70AgeqT
su4jOOKag9gxnvcEg6UAC9STWNLomCIjI39ZLGhg8wlSbE5o63s9CEizlNtiZLUX
ZwIDAQAB
-----END PUBLIC KEY-----"""

@pytest.fixture(scope="session")
def internal_cert():
    return """-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAw6VojppAkc+FOC9kLLDt
pvDkE6lAJj1Af8sgzDuTlP9iVgUzC/qYIMupsMRUzbWP1LnYKGvILQxfna5py9hK
C/BwlYn8tZ1l8gfvjnJ86DrKwLfkiJImxQPAH4ZsIXeMdnOump71u62TiQJdnZtB
1pOjiDIsAQBIiCKJEZMwIYvmT8LUaazzTsf0BzvT1Vd5swUJTuJQBXV7sT5TvGXv
o9CwGoNyspr37/86BNQQou/xEIxVd27Hpshr/98ufv7oOVDKWQA+7XFANzoOf9Rs
24LINih3fnynoVzuTt/cNPHKZAKSccV2wz83B7lA03kqezV0ihJAiOfNmbstARl5
pwIDAQAB
-----END PUBLIC KEY-----"""

@pytest.fixture(scope="session")
def external_key():
    return """-----BEGIN RSA PRIVATE KEY-----
MIIEowIBAAKCAQEAqvZ8i1qxb5ijA6Z/oS1DU9jZM4ZXRLEbErOt541cZC4/AbFj
xRfc4vTNWux9JYIjRQ761XJnPNzRvOMnh2Dm3s8O1MVC2QS4TqRz3ga2QjmqA9IS
LmV3h01K9EqeCVlYLjR0yMpUaSJcwVD7kXJCkATl4svjjpSlKNu/THd8Ybzq03Wg
MTB5SwysIWXO4jREMqvXGgxVGs54VLullTZypsWvsm7GOX+wZfTiGk3YoWCI0xYB
KWD5IRn5ykk1vt6Hr8m3qMLN8Hjn+70AgeqTsu4jOOKag9gxnvcEg6UAC9STWNLo
mCIjI39ZLGhg8wlSbE5o63s9CEizlNtiZLUXZwIDAQABAoIBADj1CaU7v+WYCqdK
rza5Z7blhedemHBwOL7QUd3VALT3m1IrETw2qQaL3GawZzk51BpL1VGLXc2vG8k8
PI0jwYb1oNFOLukyY1z/+QI9ZjdGEthXAmumrl3LaG9e9KvUskYif83tapXkNtbI
7icU1Aw7NYOLvfJgn22vWFpXbH768Af5Wizy+xEEDi2b96+y7XKCYqruOuM9aF/7
/P8zKLUSyNKfRp1fIBeGCwjJPXUjGFXnWiVu/hTOpo7faaIX4XqoRe/btnaAT6pD
eaR48I0Op4cn7qY1ocHqNnOVTbE9d4A+rfyhLn3zk15WAFIDfTVZi0Let5LCGYs8
/Y9aK3ECgYEA28o7klmMwWHbJCZ3Q1LJnFqs1He1AcNB+fSqXOHjiGoPEjUzz5xb
lND7OIUFrP6Bew7Zle/ra8tx9Qb7nbEustfv9Z1lfM3vYUO5ELD7MORgE5bp63cW
gnaR0Wt0uowQE1faa3zWSlDj1KWRMxg1Mlana61nBcbjZi/zBSlEBe0CgYEAxyDx
DbK3WHiHM8bBLaXxzlrM4nIeOk3QUInNvMktWYbL+qNhvA1rzLaNt5Km/3/gv5La
zBv4OCTxWmgbkeRZzLMnoJqHYcIaZXXfu0odu+RlWHg2weDVAdK76YMHMjeTb0Oc
zDvgqzkm7xVB/BIsDBh7G5TdzrMVYd+dDtYAaCMCgYA+jl/INr6uJ8j0kBIv0jdZ
ziEU1AFrwARePTnj99MH94/UkdnuREV4pMZr/shvxXxa5w4c881gv+XuZI9nAZAW
DQnaMXM6Q27XkyvwST8lWkrSKIjpoNqCRXcZUOlHUygcKUwcfcqMGlvOIVWjYkd5
zVoXdqI07be5VPppRKZ8vQKBgCx4Yt0YSHHtmOQrQwEaykG3bYI00iNbNWaHH/dF
Rmdz/yHeUvlYVfbF3ihdDAaLXHnsJgjnRivjtDLohR+zy4o3iGYJnKpsWtDF3k1J
ZBKOE0QKtM/fkjJyCZ1Asd65Iqf6hY3m4WnbvaIvMSlGNagN35BtUovS2rec5Mpn
aHfZAoGBANovb+N7Hw0I6VzZ3zC2O1fHvqcjvbsxtdqxhefgeLjE8E85veCHLSzY
VGlh2Jm4V9fYAJ+hWXquT8iZTUQ3fBgpi3c7zFpypqket8xBXSCK8zqiqoL5jPNs
Ja7Wtex7Q9fFHu7uGbz8Ab+7b6BDBbyCN4hZaCKREcenuplLSqXM
-----END RSA PRIVATE KEY-----"""

@pytest.fixture(scope="session")
def internal_key():
    return """-----BEGIN RSA PRIVATE KEY-----
MIIEpQIBAAKCAQEAw6VojppAkc+FOC9kLLDtpvDkE6lAJj1Af8sgzDuTlP9iVgUz
C/qYIMupsMRUzbWP1LnYKGvILQxfna5py9hKC/BwlYn8tZ1l8gfvjnJ86DrKwLfk
iJImxQPAH4ZsIXeMdnOump71u62TiQJdnZtB1pOjiDIsAQBIiCKJEZMwIYvmT8LU
aazzTsf0BzvT1Vd5swUJTuJQBXV7sT5TvGXvo9CwGoNyspr37/86BNQQou/xEIxV
d27Hpshr/98ufv7oOVDKWQA+7XFANzoOf9Rs24LINih3fnynoVzuTt/cNPHKZAKS
ccV2wz83B7lA03kqezV0ihJAiOfNmbstARl5pwIDAQABAoIBAHkZy6xpQop+v2FV
xFX9dj0dYq/g1bpVs7TmkPiZ7/IMWlwQf9ZsWPoD1pd6D1v9hHgSSQBMJu3reMvr
h8ULrlnRjH6jmO757H5x9xBfQX2l1paPvT/j52ePuL5KwGe+zg0L8gn3VvfCq03l
TvkdfxVI8bJ+C8ra2AqcBegBCYvZLVHg4jThgnbNq3FEqq0tSLgERdWLn1g3Fmal
wJpI3sRdDiRIASs4OUv/SYDpxbr2H+kdAHXidzijLtlF4guryxJqmnxe+4s2UKvB
rEe7ZH9xHeHWm5YeVLAya+7DqPsuLaxURnsSCY14oi9H/KAJwWtDr7mnT0U70Hnx
svKUJIECgYEA7D8WqtXc71zir2gbu3OK2SWVcuDk1abL1ZI3OyboZSqqPkS5dHna
uZGJpYXeNS7O4OZw5ettvozp6maG9CSx55j46cbo3Vl9niuyjdUKImHRPwl3hf8q
+gRyIFtYtHaMM2UzkVuK3nDOwsps/ACHajq/dbU510vkvyRRmEAVF8cCgYEA1AFC
r0185RvcZMZnCl+NjwNdMvyaVMv/zHq1aj6FBM4Eg+ATFkusnTRwXt3G5suvTQeb
a2OL7VosNPV9KZgljkzhRKb7rX6VpGEp04fLCLzhMtNUGpnGJ/Ls0rb12/WiJ5ID
LRBeiW6PAAJrAKfW/z0ib3tNdMiSlsV4CiGdTyECgYEA3pO91mwxgG7Bv8LVRfFg
4p8PQ7Fwx34jUu80DA3nK9Fbndj/5SYdFXJx/bm3FeIo1SknOudpePqoM8gu43xj
BLFR6mcV4925hTjkO854gtYn3z0bF+rFYdKod6W4WCXMh2uUfyGBH3umwU0YCwCw
lDwPQGXivd3qAM/F9CyD8ucCgYEA07Q+hxf19EOSHgARvUYV8g4F73hYFP1xfxu6
NVIhOt8PmzETXlxLDuDipZkJmkcj0uQZy28ot+gn5OqQBg0s06DqAshM4r9ZZ0Cv
p9Aea6dRqpNllPqh6hqnavPRJ8luF92y1jlWbh69JeFEQO9Zvp+p13Fls4zm9TBT
4qoWkIECgYEA1ZF0IVABttkUfYV+YG9JLcFi+eHbeCctlEnWZAisd2CYWKfY6rfI
XKmKPAguhan7QVFfvhXpT0RVRrJGRp65WJnwBNHn3e1FYCIz71J+wO1jMElrCQWk
gQi1+eorH63mzt6/HAZKVCnK3s2DvM87rDVY1GLCTrTNqcSICC9BvG4=
-----END RSA PRIVATE KEY-----"""

@pytest.fixture()
def pao_group(configuration, httpretty, external_cert): # pylint: disable=redefined-outer-name
    group = {
        'uri'         : '{}/groups/{}/'.format(configuration.pao_root, uuid4()),
        'name'        : 'The Sports Sesh',
        'description' : 'Sports talk show',
        'secret'      : external_cert,
    }

    httpretty.register_uri(
        httpretty.GET,
        group["uri"],
        status=200,
        body=JSON.dumps(group),
    )

    return group

@pytest.fixture
def config_override(app):
    @contextlib.contextmanager
    def _config_override(name, value): # pylint: disable=redefined-outer-name
        if name not in app.config:
            app.config[name] = value
            yield
            del app.config[name]
        else:
            previous = app.config[name]
            app.config[name] = value
            yield
            app.config[name] = previous
    return _config_override

def _uses_privileged_auth(authorization):
    if not authorization:
        return False
    parts = authorization.split(' ')
    decoded = base64.b64decode(parts[1])
    value = decoded.decode('utf-8')
    return value.split(':')[0] == 'api'

@pytest.yield_fixture
def backend_app(app):
    _backend_app = sepiida.backend.create('backend_app_fixture', testing=True, flask_app=app)
    yield _backend_app
    _backend_app.close()
    # this is needed to ensure we probably deregister our tasks on cleanup of the fixture
    # and prevent state bleed-over
    for instance in sepiida.backend.TaskProxy.instances:
        instance.celery_task = None
    sepiida.backend.create.application = None
    sepiida.backend.create.flask_app = None

@pytest.fixture
def mock_task_factory(mocker):
    def _factory(module_name):
        @sepiida.backend.task()
        def _inner(*args, **kwargs):
            _inner.ran = True
            _inner.task_args = args
            _inner.task_kwargs = kwargs
        _inner.ran = False
        _inner.task_args = None
        _inner.task_kwargs = None
        mocker.patch(module_name, side_effect=_inner)
        return _inner
    return _factory

@pytest.yield_fixture
def storage(configuration, httpretty): # pylint: disable=redefined-outer-name
    MockFile = collections.namedtuple('MockFile', ('bucket', 'content', 'key', 'mimetype', 'uuid'))
    class MockStorage():
        def __init__(self):
            self.files = {}

            url = urllib.parse.urljoin(configuration.sepiida.storage, 'file/')
            httpretty.register_uri(
                httpretty.GET,
                url,
                body=self.capture_list,
            )
            httpretty.register_uri(
                httpretty.POST,
                url,
                body=self.capture_post_file,
            )
            pattern = re.compile(url + '[\\-\\w]+/')
            httpretty.register_uri(
                httpretty.GET,
                pattern,
                body=self.capture_get,
            )
            url = urllib.parse.urljoin(configuration.sepiida.storage, 'upload/')
            pattern = re.compile(url + '[\\-\\w]+/')
            httpretty.register_uri(
                httpretty.GET,
                pattern,
                body=self.capture_get_upload,
            )
            httpretty.register_uri(
                httpretty.PUT,
                pattern,
                body=self.capture_put_upload,
            )

        def store(self, key, bucket, content, mimetype, uuid=None):
            key = str(key)
            for uuid_, file_ in self.files.items():
                if file_.key == key and file_.bucket == bucket:
                    LOGGER.debug("Overwriting storage content for %s:%s", bucket, key)
                    uuid = uuid_
                    break
            uuid = uuid or uuid4()
            uuid = str(uuid)
            self.files[uuid] = MockFile(
                bucket      = bucket,
                content     = content,
                key         = key,
                mimetype    = mimetype,
                uuid        = uuid
            )
            return uuid

        def get(self, key):
            skey = str(key)
            for file_ in self.files.values():
                if skey == file_.key:
                    return self._to_resource(file_)
            return None

        @staticmethod
        def _to_resource(file_):
            return {
                'bucket'            : file_.bucket,
                'content'           : urllib.parse.urljoin(configuration.sepiida.storage, 'upload/{}/'.format(file_.uuid)),
                'key'               : file_.key,
                'mimetype'          : file_.mimetype,
                'upload-location'   : urllib.parse.urljoin(configuration.sepiida.storage, 'upload/{}/'.format(file_.uuid)),
                'uri'               : urllib.parse.urljoin(configuration.sepiida.storage, 'file/{}/'.format(file_.uuid)),
            }

        def capture_get(self, dummy_request, uri, headers):
            uuid = uri.split('/')[-2]
            try:
                file_ = self.files[uuid]
                return 200, headers, JSON.dumps(self._to_resource(file_))
            except KeyError:
                return 404, headers, ''

        def capture_list(self, request, dummy_uri, headers):
            headers['Content-Type'] = 'application/json'
            filters = _extract_filters(request)
            files = self.files.values()
            for filter_ in filters:
                files = [f for f in files if getattr(f, filter_.name) in filter_.values]
            body = {'resources' : [self._to_resource(file_) for file_ in files]}
            return 200, headers, JSON.dumps(body)

        def capture_post_file(self, request, dummy_uri, headers):
            body = JSON.loads(request.body.decode('utf-8'))
            uuid = self.store(
                key         = body['key'],
                bucket      = body['bucket'],
                content     = None,
                mimetype    = None,
            )
            headers['Location'] = urllib.parse.urljoin(configuration.sepiida.storage, 'file/{}/'.format(uuid))
            headers['Upload-Location'] = urllib.parse.urljoin(configuration.sepiida.storage, 'upload/{}/'.format(uuid))
            LOGGER.debug("Captured POST of file %s with headers %s", self.files[uuid], headers.items())
            return 201, headers, ''

        def capture_get_upload(self, dummy_request, uri, headers):
            uuid = uri.split('/')[-2]
            file_ = self.files[uuid]
            LOGGER.debug("Captured GET of content %s which returned %s", uri, file_)
            if file_.content is None:
                return 404, headers, 'not here bro'
            return 200, headers, file_.content

        def capture_put_upload(self, request, uri, headers):
            uuid = uri.split('/')[-2]
            file_ = self.files[uuid]
            self.store(
                key         = file_.key,
                bucket      = file_.bucket,
                content     = request.body,
                mimetype    = request.headers.get('Content-Type'),
                uuid        = file_.uuid,
            )
            LOGGER.debug("Captured PUT of file content of len %d to %s", len(request.body), uri)
            return 204, headers, ''

    storage_ = MockStorage()
    yield storage_

@pytest.fixture
def disable_session_reuse():
    @contextlib.contextmanager
    def _disable_session_reuse():
        privileged_session_reuse = sepiida.requests.privileged_session.reuse
        user_session_reuse = sepiida.requests.user_session.reuse
        sepiida.requests.privileged_session.reuse = False
        sepiida.requests.user_session.reuse = False
        yield
        sepiida.requests.privileged_session.reuse = privileged_session_reuse
        sepiida.requests.user_session.reuse = user_session_reuse
    return _disable_session_reuse



import logging
import random
import time

import sepiida.backend

LOGGER = logging.getLogger(__name__)

@sepiida.backend.task(bind=True)
def do_sum(self, x, y): # pylint: disable=unused-argument
    LOGGER.info("Summing %d and %d", x, y)
    return sum([x, y])

def send_jobs():
    try:
        while True:
            x = random.randint(0, 100)
            y = random.randint(0, 100)
            LOGGER.info("Sending %d and %d", x, y)
            do_sum.delay(x=x, y=y)
            time.sleep(3)
    except KeyboardInterrupt:
        LOGGER.info("Shutting down")

APP = sepiida.backend.create('sepiida.testing',
    rabbit_url = 'amqp://guest:guest@localhost',
    flask_app = None,
)



import collections
import logging
import re
import threading
import time
from urllib import parse

import flask
import requests
import requests.auth
import sepiida.context
import sepiida.errors
import sepiida.session

LOGGER = logging.getLogger(__name__)

Request = collections.namedtuple('Request', ('method', 'url', 'params', 'data'))

class TrackedSession(requests.Session):
    "A requests Session that tracks requests that are made"
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.delay = 0
        self.requests = []

    def cleanup(self):
        LOGGER.info("Made %d %s requests with %s seconds total delay", len(self.requests), self.__class__.__name__, self.delay)

    def request(self, method, url, *args, params=None, data=None, **kwargs):
        self.requests.append(Request(
            data    = data,
            method  = method,
            params  = params,
            url     = url,
        ))
        start = time.time()
        try:
            return super().request(method, url, *args, params=params, data=data, **kwargs)
        finally:
            self.delay += time.time() - start

class PrivilegedSession(TrackedSession):
    pass

class UserSession(TrackedSession):
    pass

SESSIONS = threading.local()
def add_request_handler(app):
    "Adds hooks to an app to keep track of the external requests it makes"
    def _clear_user_session():
        SESSIONS.privileged_session          = None
        SESSIONS.user_session                = None
    def _show_session_usage(response):
        if SESSIONS.privileged_session:
            SESSIONS.privileged_session.cleanup()
        if SESSIONS.user_session:
            SESSIONS.user_session.cleanup()
        return response
    app.before_request(_clear_user_session)
    app.after_request(_show_session_usage)
    user_session.reuse = True
    privileged_session.reuse = True

def remove_request_handler():
    "Only used for test cleanup, this just undoes all the work of add_request_handler"
    user_session.reuse = False
    privileged_session.reuse = False
    if hasattr(SESSIONS, 'privileged_session'):
        del SESSIONS.privileged_session
    if hasattr(SESSIONS, 'user_session'):
        del SESSIONS.user_session

def user_session(username=None, password=None, session=None, user=None):
    current_session = getattr(SESSIONS, 'user_session', None)
    if current_session and user_session.reuse:
        return current_session
    if not any([username, password, session, user]):
        if flask.has_request_context():
            credentials = sepiida.context.extract_credentials(flask.request)
            password    = credentials['password']
            session     = credentials['session']
            user        = credentials['user']
            username    = credentials['username']
        elif sepiida.backend.has_task_context():
            credentials = sepiida.backend.task_credentials()
            password    = credentials['password']
            session     = credentials['session']
            user        = credentials['user']
            username    = credentials['username']
        else:
            msg = ("You have attempted to create a user session without having a flask request context or a backend task context. "
                   "You must therefore supply your own credentials as parameters")
            raise Exception(msg)

    _user_session = UserSession()
    if username and password:
        LOGGER.debug("Using credentials %s %s for user session", username, password)
        _user_session.auth = requests.auth.HTTPBasicAuth(username, password)
    elif session:
        LOGGER.debug("Passing through session token")
        _user_session.cookies['session'] = session
    else:
        LOGGER.warning("User session requested, but I have no session data to use")
    _user_session.headers.update({'Request-ID': sepiida.session.current_request_id()})
    if hasattr(SESSIONS, 'user_session'):
        SESSIONS.user_session = _user_session
    return _user_session
user_session.reuse = False

class PrivilegedAuth(requests.auth.HTTPBasicAuth):
    def __call__(self, prepared_request):
        prepared_request = super(self.__class__, self).__call__(prepared_request)
        parts = parse.urlsplit(prepared_request.url)
        domains = flask.current_app.config.get('SEPIIDA_INTERNAL_DOMAINS', [])
        domains = '|'.join([re.escape(domain) for domain in domains])
        domains_regex = r'(.*)?\.({0})$'.format(domains)
        if not parts.scheme == 'https':
            raise InvalidPrivilegedScheme()
        if not re.fullmatch(domains_regex, parts.netloc):
            raise InvalidPrivilegedDomain()

        return prepared_request

def privileged_session(api_token=None):
    current_session = getattr(SESSIONS, 'privileged_session', None)
    if current_session and privileged_session.reuse:
        return SESSIONS.privileged_session
    if not api_token:
        api_token = flask.current_app.config['API_TOKEN']

    _session = PrivilegedSession()
    _session.auth = PrivilegedAuth('api', api_token)
    _session.headers.update({'Request-ID': sepiida.session.current_request_id()})
    if hasattr(SESSIONS, 'privileged_session'):
        SESSIONS.privileged_session = _session
    return _session
privileged_session.reuse = False

class InvalidPrivilegedDomain(sepiida.errors.Error):
    def __init__(self):
        super(
            self.__class__,
            self,
        ).__init__(
            "Privileged domain must be trusted",
            "invalid-privileged-domain"
        )

class InvalidPrivilegedScheme(sepiida.errors.Error):
    def __init__(self):
        super(
            self.__class__,
            self,
        ).__init__(
            "Privileged request requires https",
            "invalid-privileged-sheme"
        )



import logging

import flask
from raven.contrib.flask import Sentry
from werkzeug.contrib.fixers import ProxyFix

LOGGER = logging.getLogger(__name__)

def handle_reverse_proxy(app):
    LOGGER.info("Adding support for nginx reverse proxy fix in sentry")
    app.wsgi_app = ProxyFix(app.wsgi_app)

class SepiidaSentry(Sentry):
    def get_user_info(self, request):
        return flask.g.current_user

def add_sentry_support(application, sentry_dsn, version):
    raven = SepiidaSentry(app=application, dsn=sentry_dsn, logging=True, level=logging.ERROR)
    raven.client.release = version
    application.config.setdefault('RAVEN', raven)
    logging.getLogger('webargs.flaskparser').setLevel(logging.CRITICAL)
    return raven



import inspect
import json
import logging
import math
import urllib.parse
import uuid
from datetime import date, datetime, timezone

import arrow
import flask

import sepiida.routing
from sepiida.errors import APIException, AppError, Error

LOGGER = logging.getLogger(__name__)

class UnpackageError(Exception):
    def __init__(self, errors):
        super().__init__()
        assert isinstance(errors, list)
        self.errors = errors

class PackageError(Exception):
    def __init__(self, errors):
        super().__init__()
        assert isinstance(errors, list)
        self.errors = errors
    def __str__(self):
        return "\n".join(self.errors)

class FakeArguments():
    def __init__(self, fields):
        self.fields = FakeFieldsArgument(fields)

class FakeFieldsArgument():
    def __init__(self, values):
        self.values = values

TYPES_TO_NAME = {
    date        : 'a date',
    datetime    : 'a datetime',
    dict        : 'an object',
    list        : 'an array',
    int         : 'a number',
    float       : 'a number',
    bool        : 'a boolean',
    str         : 'a string',
    type(None)  : 'null',
    uuid.UUID   : 'a uuid',
    'URI'       : 'a uri',
    'URL'       : 'a URL',
}
def _unpackage_error_message(name, expected_type, provided_data):
    provided_type_name = TYPES_TO_NAME[type(provided_data)]
    expected_type_name = TYPES_TO_NAME[expected_type]
    return "The property '{}' must be {}. You provided {}".format(name, expected_type_name, provided_type_name)

def _raise_validation_error(title):
    errors = [Error('request-validation-error', title)]
    raise UnpackageError(errors)

def _raise_validation_type_error(name, expected_type, provided_data):
    title = _unpackage_error_message(name, expected_type, provided_data)
    errors = [Error('type-validation-error', title)]
    raise UnpackageError(errors)

def _maybe_raise_all_sub_errors(errors):
    if not errors:
        return

    sub_errors = []
    for e in errors:
        sub_errors += e.errors
    raise UnpackageError(errors=sub_errors)

class Unspecified():
    pass

class Field(): # pylint: disable=too-many-instance-attributes
    TYPE_NAME = 'raw-field'
    def __init__(self, # pylint: disable=too-many-arguments
            choices    = None,
            default    = Unspecified,
            docs       = None,
            example    = Unspecified,
            filterable = True,
            methods    = None,
            nullable   = False,
            optional   = False,
            rename     = None,
            sortable   = True,
        ):
        self.choices = choices
        if default != Unspecified:
            self.default = default
        self.docs           = docs
        self.filterable     = filterable
        self.methods        = methods if methods is not None else ['GET', 'POST', 'PUT']
        self.nullable       = nullable
        self.optional       = optional
        self.rename         = rename
        self.sortable       = sortable

        self._generate_example(example, choices)

    def _check_choices(self, name, data):
        if self.choices is None:
            return

        if data not in self.choices:
            choices = ["'{}'".format(choice) for choice in self.choices]
            choices = ", ".join(choices)
            error = Error(
                'property-value-not-in-allowed-values',
                "The property '{}' must be one of {}. You provided '{}'".format(name, choices, data)
            )
            raise UnpackageError([error])

    def _generate_example(self, example, choices):
        if example != Unspecified:
            self.example = example
        elif choices:
            self.example = '|'.join([str(choice) for choice in choices])

    def package(self, data):
        raise NotImplementedError

    def unpackage(self, name, package, method):
        raise NotImplementedError

    def get_options_description(self):
        description = {
            'type'          : self.TYPE_NAME,
            'choices'       : self.choices,
            'filterable'    : self.filterable,
            'methods'       : self.methods,
            'optional'      : self.optional,
            'sortable'      : self.sortable,
        }
        if hasattr(self, 'default'):
            description['default'] = self.default
        return description

    def get_filter_value(self, name, value): # pylint: disable=unused-argument,no-self-use
        return value

    def process_arguments(self, arguments):
        pass

    def schema(self, method='POST'):
        my_schema = {
            'description'   : self.docs,
            'filterable'    : self.filterable,
            'nullable'      : self.nullable,
            'type'          : self.TYPE_NAME,
            'sortable'      : self.sortable,
        }
        if method == 'GET':
            my_schema['optional'] = self.optional
        if method == 'POST' and hasattr(self, 'default'):
            my_schema['default'] = self.default
        if self.choices:
            my_schema['enum'] = self.choices
        if hasattr(self, 'example'):
            my_schema['example'] = json.dumps(self.example)
        return my_schema

class Anything(Field):
    TYPE_NAME = 'anything'
    def package(self, response):
        return response

    def unpackage(self, name, data, method):
        return data

    def schema(self, method='POST'):
        return super().schema(method)

class SignatureField(Field): # pylint: disable=abstract-method
    def __init__(self, s=None, **kwargs):
        super().__init__(**kwargs)
        self.signature = s
        try:
            for k, v in self.signature.items():
                if inspect.isclass(v):
                    raise Exception(("You have provided a class '{}' as part of the signature for {}."
                        "Please specify the signature using instances of the class, not the class itself".format(v, k)))
        except AttributeError:
            pass

    def _process_against_signature(self, arguments):
        errors = []
        if arguments.fields and not self.signature:
            errors.append(Error(
                status_code=400,
                error_code='unable-to-specify-fields',
                title=("You specified return fields. This resource cannot return a subset of its fields"),
            ))

        if arguments.filters and not self.signature:
            errors.append(Error(
                status_code=400,
                error_code='unable-to-specify-filters',
                title=("You specified filters. This resource cannot be filtered"),
            ))
        return errors

    def _process_fields(self, arguments):
        if not isinstance(self.signature, SignatureField):
            return []
        return self.signature._process_fields(arguments) # pylint: disable=protected-access

    def _process_filters(self, arguments):
        if not isinstance(self.signature, SignatureField):
            return []
        return self.signature._process_filters(arguments) # pylint: disable=protected-access

    def _process_sorts(self, arguments):
        if not isinstance(self.signature, SignatureField):
            return []
        return self.signature._process_sorts(arguments) # pylint: disable=protected-access

    def process_arguments(self, arguments):
        errors = self._process_against_signature(arguments)
        errors += self._process_fields(arguments)
        errors += self._process_filters(arguments)
        errors += self._process_sorts(arguments)

        if errors:
            raise APIException(errors=errors)

class Object(SignatureField):
    TYPE_NAME = 'object'
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.signature:
            self.example = {k: v.example for k, v in self.signature.items() if hasattr(v, 'example')}
        else:
            self.example = None
        if self.signature:
            fieldnames = set(self.signature.keys())
            for key, field in self.signature.items():
                if field.rename and field.rename in fieldnames:
                    raise Exception((
                        "You cannot rename field {} to {} because another "
                        "field is already using that name").format(key, field.rename))

    def _fieldnames(self):
        results = list(self.signature.keys())
        for name, field in self.signature.items():
            if isinstance(field, SignatureField):
                results += ["{}.{}".format(name, f) for f in field._fieldnames()] # pylint: disable=protected-access
        return results

    def _filternames(self):
        results = [name for name, field in self.signature.items() if field.filterable]
        for name, field in self.signature.items():
            if isinstance(field, SignatureField):
                results += ["{}.{}".format(name, f) for f in field._filternames()] # pylint: disable=protected-access
        return results

    def _process_fields(self, arguments):
        if not self.signature:
            return []

        errors = []
        for field in (arguments.fields.values if arguments.fields else []):
            if field not in self.signature:
                parent_field, _, child_field = field.partition('.')
                if parent_field in self.signature:
                    self.signature[parent_field]._process_fields(FakeArguments([child_field])) # pylint: disable=protected-access
                else:
                    msg = ("You specified the resource return '{}' but it is not a valid field for this resource. "
                           "Valid fields include {}").format(field, _quote(self._fieldnames()))
                    errors.append(Error('invalid-field-specified', msg))
        return errors

    def _process_filters(self, arguments):
        if not self.signature:
            return []

        errors = []
        for _filter in arguments.filters:
            try:
                field = self.signature[_filter.name]
            except KeyError:
                if self.signature:
                    title = ("Your requested filter property, '{}', is not a valid property for this endpoint. "
                            "Please choose one of {}".format(_filter.name, _quote(self._filternames())))
                else:
                    title = "You requested a filter on property '{}'. This endpoint does not allow filtering".format(_filter.name)
                errors.append(Error('invalid-filter-property', title))
                continue

            if not field.filterable:
                title = ("Your requested filter property, '{}', is not a valid property for this endpoint. "
                         "Please choose one of {}").format(_filter.name, _quote(self._filternames()))

                errors.append(Error(
                    status_code = 400,
                    error_code  = 'invalid-filter-property',
                    title       = title,
                ))
                continue

            new_values = []
            for value in _filter.values:
                try:
                    new_value = field.get_filter_value(_filter.name, value)
                    if field.choices and new_value not in field.choices:
                        title = ("You requested a filter on '{name}' to the value '{value}'. "
                                 "The property '{name}' only permits values from a particular "
                                 "set of options. Please use one of the following options for "
                                 "your filter: {values}").format(name=_filter.name, value=value, values=', '.join(field.choices))
                        errors.append(Error(
                            status_code = 400,
                            error_code  = 'invalid-filter-value',
                            title       = title,
                        ))
                        continue
                    new_values.append(new_value)
                except ValueError:
                    title = ("You requested a filter on '{name}' to the value '{value}'. "
                            "The property '{name}' is a {type} and so filters on it must "
                            "also be {type}").format(name=_filter.name, value=value, type=field.TYPE_NAME)
                    errors.append(Error(
                        status_code = 400,
                        error_code  = 'invalid-filter-value',
                        title       = title,
                    ))
                except AppError as e:
                    errors.append(Error(
                        status_code = 500,
                        error_code  = 'app-error',
                        title       = str(e)
                    ))
            _filter.values = new_values
        return errors

    def _process_sorts(self, arguments):
        errors = []
        for sort in (arguments.sorts.values if arguments.sorts else []):
            try:
                field = self.signature[sort.name]
            except KeyError:
                if self.signature:
                    title = ("Your requested sort property, '{}', is not a valid property for this endpoint. "
                            "Please choose one of {}".format(sort.name, _quote(self._sortnames())))
                else:
                    title = "You requested a sort on property '{}'. This endpoint does not allow sorting".format(sort.name)
                errors.append(Error(status_code=400, error_code='invalid-sort-property', title=title))
                continue

            if not field.sortable:
                title = ("Your requested sort property, '{}', has sorting disabled for this endpoint. "
                         "Please remove it and try again. "
                         "The following fields are allowed for sorting: {}").format(sort.name, _quote(self._sortnames()))
                errors.append(Error(status_code=400, error_code='disabled-sort-property', title=title))
                continue

        return errors

    def _sortnames(self):
        return [name for name, field in self.signature.items() if field.sortable]

    def package(self, response):
        package = {}
        if response is None:
            return None

        if self.signature is None:
            return dict(response)

        if response is None:
            return None

        errors = []
        for k, v in self.signature.items():
            try:
                if 'GET' in v.methods:
                    try:
                        newkey = k if v.rename is None else v.rename
                        package[k] = v.package(response[newkey])
                    except (TypeError, ValueError) as e:
                        errors.append("Programmer failed to provide the correct type for {}: {}".format(k, e))
            except KeyError:
                if not v.optional:
                    errors.append("Programmer failed to provide {}, which is a required value to return for this resource.".format(k))
        if errors:
            raise PackageError(errors=errors)
        return package

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, dict):
            raise _raise_validation_type_error(name, dict, data)

        self._check_choices(name, data)

        if self.signature is None:
            return data

        package = {}
        errors = []
        for k, v in data.items():
            try:
                sub_name = "{}{}".format(name + "." if name else '', k)
                try:
                    sub_signature = self.signature[k]
                except KeyError:
                    raise _raise_validation_error("You provided '{}'. It is not a valid property for this resource".format(sub_name))
                value = sub_signature.unpackage(sub_name, v, method)
            except UnpackageError as e:
                errors.append(e)
                value = None
            package[k] = value
        for k, v in self.signature.items():
            sub_name = "{}{}".format(name + "." if name else '', k)
            if method not in v.methods and k in package:
                post_errors = [Error('cannot-set-property', "You provided '{}'. It cannot be directly set via {}".format(sub_name, method))]
                errors.append(UnpackageError(post_errors))
            if method == 'POST' and method in v.methods and k not in package:
                try:
                    package[k] = getattr(v, 'default')
                except AttributeError:
                    post_errors = [Error('missing-parameter',
                        "You did not provide the {} property. It is required when POSTing to this resource".format(sub_name)
                    )]
                    errors.append(UnpackageError(post_errors))
            if v.rename is not None:
                package[v.rename] = package.pop(k)
        _maybe_raise_all_sub_errors(errors)
        return package

    def get_options_description(self):
        base = super().get_options_description()
        if self.signature:
            base['fields'] = {k: v.get_options_description() for k, v in self.signature.items()}
        return base

    def schema(self, method='POST'):
        my_schema = super().schema(method)
        if not self.signature:
            return my_schema
        my_schema.update({
            'properties'    : {k: v.schema(method) for k, v in self.signature.items()},
        })
        if method == 'POST':
            my_schema['required'] = [k for k, v in self.signature.items() if not hasattr(v, 'default')]
        elif method == 'GET':
            my_schema['required'] = [k for k, v in self.signature.items() if not v.optional]
        return my_schema

class ObjectEnvelope(Object):
    """
    This class is essentially for sepiida internal use. It's designed to be more lenient than the regular Object
    class for use in the LIST part of endpoint. Older versions of sepiida did not enforce that various queries
    worked differently on a LIST vs a GET. This class papers over that design choice from long ago to maintain
    backwards compatibility
    """
    def __init__(self, s):
        signature = {
            'resources' : Array(s=s)
        }
        super().__init__(s=signature)

    def process_arguments(self, arguments):
        errors = self._process_against_signature(arguments)
        errors += super()._process_fields(arguments)
        errors += self.signature['resources']._process_filters(arguments) # pylint: disable=protected-access
        errors += self.signature['resources']._process_sorts(arguments) # pylint: disable=protected-access

        if errors:
            raise APIException(errors=errors)

    def _filternames(self):
        return self.signature['resources']._filternames() # pylint: disable=protected-access

    def _sortnames(self):
        return self.signature['resources']._sortnames() # pylint: disable=protected-access

    def package(self, response):
        return super().package({
            'resources' : response
        })

class Array(SignatureField):
    TYPE_NAME = 'array'
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.example  = [self.signature.example if self.signature and hasattr(self.signature, 'example') else '...']

    def _fieldnames(self):
        return self.signature._fieldnames() if isinstance(self.signature, SignatureField) else [] # pylint: disable=protected-access

    def _filternames(self):
        return self.signature._filternames() if isinstance(self.signature, SignatureField) else [] # pylint: disable=protected-access

    def package(self, response):
        if response is None:
            return None

        _convert = (lambda x: self.signature.package(x)) if self.signature else (lambda x: x) # pylint: disable=unnecessary-lambda
        return [_convert(value) for value in response]

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, list):
            raise _raise_validation_type_error(name, list, data)

        self._check_choices(name, data)

        if self.signature is None:
            return data

        package = []
        errors = []
        for i, d in enumerate(data):
            try:
                sub_name = "{}{}".format(name + "." if name else '', i)
                value = self.signature.unpackage(sub_name, d, method)
            except UnpackageError as e:
                errors.append(e)
                value = None
            package.append(value)
        _maybe_raise_all_sub_errors(errors)
        return package

    def schema(self, method='POST'):
        my_schema = super().schema(method)
        if not self.signature:
            return my_schema
        my_schema.update({
            'items'    : self.signature.schema(method=method)
        })
        return my_schema
# Backwards compatibility with previous sepiida versions
ist = Array

class String(Field):
    TYPE_NAME = 'string'
    def package(self, response):
        return str(response) if response is not None else None

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, str):
            raise _raise_validation_type_error(name, str, data)

        self._check_choices(name, data)

        return data

class Integer(Field):
    TYPE_NAME = 'integer'
    def package(self, response):
        return int(response) if response is not None else None

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, int) or isinstance(data, bool):
            raise _raise_validation_type_error(name, int, data)

        self._check_choices(name, data)

        return int(data)

    def get_filter_value(self, name, value):
        parsed = float(value)
        if int(parsed) != parsed:
            raise ValueError("It's a float but not an int")
        return int(parsed)

class Float(Field):
    TYPE_NAME = 'float'
    def package(self, response):
        if isinstance(response, float) and any([math.isnan(response), math.isinf(response)]):
            raise ValueError("Programmer provided '{}' which cannot be serialized correctly".format(response))
        return float(response) if response is not None else None

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        acceptable_types = [isinstance(data, x) for x in (int, float)]
        unacceptable_types = [isinstance(data, bool)]
        if not any(acceptable_types) or any(unacceptable_types):
            raise _raise_validation_type_error(name, float, data)

        self._check_choices(name, data)

        return float(data)

    def get_filter_value(self, name, value):
        return float(value)

class Boolean(Field):
    TYPE_NAME = 'boolean'
    def package(self, response):
        return bool(response) if response is not None else None

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, bool):
            raise _raise_validation_type_error(name, bool, data)

        self._check_choices(name, data)

        return bool(data)

    def get_filter_value(self, name, value):
        lower = value.lower()
        if lower in ('f', 'false') or lower == '0':
            return False
        elif lower in ('t', 'true') or lower == '1':
            return True
        else:
            raise ValueError("{} does not parse to a bool".format(value))

class Date(Field):
    TYPE_NAME = 'iso-8601 date'
    def package(self, response):
        if response is None:
            return None

        if isinstance(response, date):
            return response.isoformat()
        else:
            raise TypeError("{} is not a valid date".format(response))

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, str):
            raise _raise_validation_type_error(name, date, data)

        try:
            return arrow.get(data).date()
        except arrow.parser.ParserError:
            raise _raise_validation_type_error(name, date, data)

    def get_filter_value(self, name, value):
        try:
            return arrow.get(value).date()
        except arrow.parser.ParserError:
            raise ValueError("{} does not parse to a date".format(value))

class Datetime(Field):
    TYPE_NAME = 'iso-8601 datetime'
    def package(self, response):
        if response is None:
            return None

        if isinstance(response, datetime):
            if response.tzinfo == timezone.utc:
                return response.isoformat()
            if response.tzinfo is None:
                return response.replace(tzinfo=timezone.utc).isoformat()
            else:
                return response.astimezone(tz=timezone.utc).isoformat()
        else:
            raise TypeError("{} is not a valid datetime".format(response))

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, str):
            raise _raise_validation_type_error(name, datetime, data)

        try:
            return arrow.get(data).datetime.astimezone(tz=timezone.utc)
        except arrow.parser.ParserError:
            raise _raise_validation_type_error(name, datetime, data)

    def get_filter_value(self, name, value):
        try:
            return arrow.get(value).datetime
        except arrow.parser.ParserError:
            raise ValueError("{} does not parse to a datetime".format(value))

class UUID(Field):
    TYPE_NAME = 'uuid'
    def package(self, response):
        if response is None:
            return None

        if isinstance(response, uuid.UUID):
            return str(response)
        else:
            raise ValueError("{} is not a valid UUID".format(response))

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, str):
            raise _raise_validation_type_error(name, uuid.UUID, data)

        try:
            return uuid.UUID(data)
        except ValueError:
            raise _raise_validation_type_error(name, uuid.UUID, data)

    def get_filter_value(self, name, value):
        return uuid.UUID(value)

def _app_has_endpoint(app, endpoint):
    for rule in app.url_map.iter_rules():
        if rule.endpoint == endpoint:
            return True
    return False

class URI(String):
    TYPE_NAME = 'URI'
    def __init__(self, endpoint, **kwargs):
        super().__init__(**kwargs)
        self.endpoint = endpoint

    def package(self, response):
        if response is None:
            return None
        elif isinstance(response, (int, uuid.UUID)):
            return sepiida.routing.uri(self.endpoint, response)
        elif isinstance(response, str):
            endpoint, _ = sepiida.routing.extract_parameters(flask.current_app, method='GET', url=response)
            assert endpoint == self.endpoint, "URI is valid but for the wrong endpoint"
            return response
        else:
            return sepiida.routing.uri(self.endpoint, **response)

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None
        try:
            route, parameters = sepiida.routing.extract_parameters(flask.current_app, method='GET', url=data)
            if not route:
                raise _raise_validation_error("The URI you provided for {} - {} - is not recognized as a URI".format(name, data))
            if route != self.endpoint:
                msg = (
                    "The URI you provided for {property_name} - {data} - "
                    "appears to be a URI for a {bad_route} resource, not a {good_route} resource".format(
                        property_name   = name,
                        data            = data,
                        bad_route       = route,
                        good_route      = self.endpoint,
                ))
                raise _raise_validation_error(msg)
        except AttributeError:
            raise _raise_validation_type_error(name, 'URI', data)
        parameters = parameters or {}
        parameters['uri'] = data
        return parameters

    def get_filter_value(self, name, value):
        _, params = sepiida.routing.extract_parameters(flask.current_app, method='GET', url=value)
        _uuid = params.get('uuid') or params.get('_id')
        if not _uuid:
            if _app_has_endpoint(flask.current_app, self.endpoint):
                raise ValueError("{} is not a recognized URI for {}".format(value, self.endpoint))
            else:
                raise AppError("The endpoint '{}' is not a valid endpoint for this application".format(self.endpoint))
        return _uuid

class URL(String):
    TYPE_NAME = 'URL'
    def package(self, response):
        if response is None:
            return response
        if not self._valid_url(response):
            message = "The URL supplied by the programmer '{}' is not a valid URL".format(response)
            LOGGER.error(message)
        return self._strip_credentials(response)

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None
        if not self._valid_url(data):
            raise _raise_validation_type_error(name, 'URL', data)
        return data

    @staticmethod
    def _valid_url(url):
        try:
            parsed = urllib.parse.urlparse(url)
            return all((
                bool(parsed.netloc),
                bool(parsed.scheme),
            ))
        except AttributeError:
            return False

    @staticmethod
    def _strip_credentials(url):
        parsed = urllib.parse.urlparse(url)
        netloc = parsed.netloc
        if parsed.username is not None:
            netloc = netloc[len(parsed.username) + 1:]
        if parsed.password is not None:
            netloc = netloc[len(parsed.password) + 1:]
        return urllib.parse.urlunparse((parsed[0], netloc, parsed[2], parsed[3], parsed[4], parsed[5]))

    def get_filter_value(self, name, value):
        if not self._valid_url(value):
            raise ValueError("{} is not a URL".format(value))
        return value

def _quote(collection):
    return ", ".join(["'{}'".format(c) for c in sorted(collection)])



import flask
import sepiida.session

def extract_credentials(_request):
    if flask.has_request_context():
        user     = sepiida.session.current_user_uri()
        username = _request.authorization['username'] if _request.authorization else None
        password = _request.authorization['password'] if _request.authorization else None
        session  =  _request.cookies.get('session', None)
    else:
        user     = None
        username = None
        password = None
        session  = None
    return {
        'password'  : password,
        'session'   : session,
        'user'      : user,
        'username'  : username,
    }



import logging
import zlib

LOGGER = logging.getLogger(__name__)

BLOCKSIZE = 200
class ZBuffer():
    ZIPPER = lambda: None
    def __init__(self, stream, blocksize=BLOCKSIZE):
        self.blocksize  = blocksize
        self.done       = False
        self.stream     = stream
        self.zipper     = self.ZIPPER()

    def __iter__(self):
        return self

    def __next__(self):
        if self.done:
            LOGGER.debug("Stopping iteration")
            raise StopIteration()
        block = self.read(self.blocksize)
        return block

    def close(self):
        self.stream.close()

    def read(self, size=None):
        result = b''
        while result == b'':
            if self.done:
                return result
            block = self.stream.read(size)
            if not block:
                LOGGER.debug("block is empty, flushing")
                self.done = True
                result = self.zipper.flush()
                LOGGER.debug("Flushed underlying zipped block and got %d bytes", len(result))
                return result
            result = self.process(block)
        LOGGER.debug("Providing zipped block of %d bytes from source of %d bytes", len(result), len(block))
        return result

    def process(self, block):
        raise NotImplementedError()

class ZipBuffer(ZBuffer):
    ZIPPER = zlib.compressobj
    def process(self, block):
        return self.zipper.compress(block)

class UnzipBuffer(ZBuffer):
    ZIPPER = zlib.decompressobj
    def process(self, block):
        return self.zipper.decompress(block)



import re

from flask_cors import CORS

DEFAULT_EXPOSE_HEADERS = [
    'Location',
    'X-Sentry-ID',
]

def register_cors_handlers(app, domains=None, supports_credentials=False, expose_headers=None):
    expose_headers = expose_headers or DEFAULT_EXPOSE_HEADERS
    if not domains:
        return

    if not isinstance(domains, list):
        domains = [domains]

    origins = '|'.join([re.escape(domain) for domain in domains])
    origin_regex = r'(.*)?\.({0})$'.format(origins)

    app.config['CORS_RESOURCES'] = {
        r'/*': { "origins": origin_regex }
    }

    return CORS(
        app,
        supports_credentials=supports_credentials,
        expose_headers=expose_headers,
    )



import attrdict
import yaml


def _parse_bool(data):
    if data in (True, 'true', '1'):
        return True
    elif data in (False, 'false', '0'):
        return False
    else:
        raise ValueError("{} must be one of true/false or 1/0".format(data))

class Option():
    SPECIAL_PARSERS = {
        bool: _parse_bool,
    }
    def __init__(self, option_type, testing, description=None):
        if not (isinstance(testing, option_type) or testing is None):
            raise Exception("You cannot specify a testing value of {} because it is not of type {}".format(testing, option_type.__name__))
        self.option_type = option_type
        self.testing     = testing
        self.description = description

    def parse(self, data):
        parser = self.SPECIAL_PARSERS.get(self.option_type, self.option_type)
        return parser(data)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "sepiida.config.Option(option_type={}, testing={})".format(self.option_type.__name__, self.testing)

def _readfile(filename):
    with open(filename, 'r') as f:
        return f.read()

def get():
    if not get.cached:
        raise Exception("You must call sepiida.config.load first")
    return get.cached
get.cached = None

def load(filename, spec):
    content = _readfile(filename)
    get.cached = parse(content, spec)
    return get.cached

def parse(content, spec):
    data = yaml.load(content)
    config = {}
    for name, option in spec.items():
        try:
            config[name] = option.parse(data[name])
        except KeyError:
            raise KeyError("The config option '{}' is required and was not provided".format(name))
        except ValueError:
            err = ("The config option '{}' was provided with value "
                   "'{}' which could not be parsed into a {}").format(name, data[name], option.option_type.__name__)
            raise ValueError(err)
    return attrdict.AttrDict(config)

def extract_test_values(spec):
    return attrdict.AttrDict({k: v.testing for k, v in spec.items()})
